/*********************************************************************************
 M E S H  Utilities
**********************************************************************************
 AUTHOR: P. Bourdin

 CREATED: April - May 2007
*********************************************************************************/

Function polyLineLength

  /****************************************************************
    Computes the arc-length of a polyline.

    Inputs:
      -- I1: First index;
      -- I2: Last  index.

    Output:
      -- length: polyline arc-length.
  ****************************************************************/

  length = 0;

  For I In {I1+1:I2}
    XYZa[] = Point{I-1};
    XYZb[] = Point{I};

    length += Sqrt( (XYZb[0]-XYZa[0])^2 +
                    (XYZb[1]-XYZa[1])^2 + (XYZb[2]-XYZa[2])^2 );
  EndFor

Return

Function Stretch

  /****************************************************************
    Computes the cell expansion ratio along a mesh line according
    to a geometric progression. Given are the first term (length
    of the first cell), the number of terms (ie number of cells)
    and the partial sum (ie overall arc-length to be discretized)
    of the corresponding geometric series.

    Inputs:
      -- DS1: arc-length of the first intervall;
      -- sumDS: overall arc-length to be discretized;
      -- N: number of cells.

    Output:
      -- R: expansion ratio.
  ****************************************************************/

  If (DS1 != sumDS)
    eps1 = 1.e-05;
    eps2 = 1.e-04;

    R = 1.5;

    For iter In {1:50}
      F  = (R - 1.)*sumDS - DS1*(R^N - 1.);
      Fp = sumDS - N*DS1*R^(N-1);

      Riter = R - F/Fp; // Newton step

      If (1. - eps2 < Riter && Riter < 1.) Riter = 1.e+01; EndIf
      If (1. < Riter && Riter < 1. + eps2) Riter = 1.e-01; EndIf

      If (iter != 50) R = Riter; EndIf
    EndFor

    If (Fabs(Riter - R) >= R*eps1)
      Printf("  no convergence in stretch");
      Printf("  N, R, Riter, sumDS, DS1: %g %g %g %g %g", N, R, Riter, sumDS, DS1);
      Printf("  proceed with equal spacing. R = 1.");
      R = 1.00001;
    EndIf

    If (Fabs(Riter - R) < R*eps1)
      R = Riter;
      DSN = DS1 * R^(N-1);
      Printf("  N, R, sumDS, DS1, DSN: %g %g %g %g %g", N, Riter, sumDS, DS1, DSN);
    EndIf
  EndIf

  If (DS1 == sumDS)
    N = 1;
    R = 1.00001;
  EndIf

Return

Function PlaneNormal

  /****************************************************************
    Computes the vector normal to a plane.

    Inputs:
      -- PP1, PP2, PP3: indices of three points defining the plane.

    Outputs:
      -- PNORMAL: plane normal (vector).
  ****************************************************************/

  xyz1[] = Point{PP1};
  xyz2[] = Point{PP2};
  xyz3[] = Point{PP3};

  xyzA[] = {};
  xyzB[] = {};

  xyzA[0] = xyz2[0] - xyz1[0];
  xyzA[1] = xyz2[1] - xyz1[1];
  xyzA[2] = xyz2[2] - xyz1[2];

  xyzB[0] = xyz3[0] - xyz1[0];
  xyzB[1] = xyz3[1] - xyz1[1];
  xyzB[2] = xyz3[2] - xyz1[2];

  PNORMAL[] = {};

  PNORMAL[0] = xyzA[1]*xyzB[2] - xyzA[2]*xyzB[1];
  PNORMAL[1] = xyzA[2]*xyzB[0] - xyzA[0]*xyzB[2];
  PNORMAL[2] = xyzA[0]*xyzB[1] - xyzA[1]*xyzB[0];

  norm = Sqrt(PNORMAL[0]^2 + PNORMAL[1]^2 + PNORMAL[2]^2);

  PNORMAL[0] /= norm;
  PNORMAL[1] /= norm;
  PNORMAL[2] /= norm;

Return

Function Inflate

  /****************************************************************
    Extrudes a 2D surface mesh (ie a graded curve) into a boundary
    layer mesh.

    Notes:
      -- it is assumed that inner point index
         increases when travelling from IB1 to IB2.

    Inputs:
      -- INFL: inflation length;
      -- IB1: index of 1st boundary point;
      -- IB2: index of 2nd boundary point;
      -- I1: index of 1st inner point;
      -- I2: index of last inner point;
      -- STRIDE: stride from I1 to I2;
      -- DIR1: direction of extrusion for 1st boundary point;
      -- DIR2: direction of extrusion for 2nd boundary point;
      -- PNORMAL: plane normal;
      -- KX, KY, KZ: switches to account for X-, Y-, Z-components;
         (KX=0 KY=1 KZ=1 corresponds to projection onto YZ plane).

    Outputs:
      -- IB1EX: extruded IB1;
      -- IB2EX: extruded IB2;
      -- IL: index of extruded curve.
  ****************************************************************/

  Geometry.AutoCoherence = 0;

  II1 = I1 + 1;
  II2 = I2 - 1;

  DIR1[0] *= KX; DIR1[1] *= KY; DIR1[2] *= KZ;
  DIR2[0] *= KX; DIR2[1] *= KY; DIR2[2] *= KZ;

  If (DIR1[0] == 0 && DIR1[1] == 0 && DIR1[2] == 0) II1 = I1; EndIf
  If (DIR2[0] == 0 && DIR2[1] == 0 && DIR2[2] == 0) II2 = I2; EndIf

  // Extrude boundary points
  If ( !(DIR1[0] == 0 && DIR1[1] == 0 && DIR1[2] == 0) )
    tmp[] = Translate {INFL*DIR1[0], INFL*DIR1[1], INFL*DIR1[2]} {
              Duplicata{
                Point{IB1};
              }
            };

    IB1EX = tmp[0];
  EndIf

  If ( !(DIR2[0] == 0 && DIR2[1] == 0 && DIR2[2] == 0) )
    tmp[] = Translate {INFL*DIR2[0], INFL*DIR2[1], INFL*DIR2[2]} {
              Duplicata{
                Point{IB2};
              }
            };

    IB2EX = tmp[0];
  EndIf

  // Extrude inner points along surface normal
  If (I2 - I1 < 2)
    Printf(" !!!! Inflate Error: not enough inner points !!!!");
  EndIf

  i = 0;
  NEWI[] = {};

  For I In {II1:II2:STRIDE}
    xyz1[] = Point{I-1};
    xyz2[] = Point{I+1};

    xyzA[0] = PNORMAL[0];
    xyzA[1] = PNORMAL[1];
    xyzA[2] = PNORMAL[2];

    xyzB[0] = xyz2[0] - xyz1[0];
    xyzB[1] = xyz2[1] - xyz1[1];
    xyzB[2] = xyz2[2] - xyz1[2];

    snormal[] = {};

    snormal[0] = xyzA[1]*xyzB[2] - xyzA[2]*xyzB[1];
    snormal[1] = xyzA[2]*xyzB[0] - xyzA[0]*xyzB[2];
    snormal[2] = xyzA[0]*xyzB[1] - xyzA[1]*xyzB[0];

    snormal[0] *= KX;
    snormal[1] *= KY;
    snormal[2] *= KZ;

    norm = Sqrt(snormal[0]^2 + snormal[1]^2 + snormal[2]^2);

    snormal[0] /= norm;
    snormal[1] /= norm;
    snormal[2] /= norm;

    tmp[] = Translate {INFL*snormal[0], INFL*snormal[1], INFL*snormal[2]} {
              Duplicata{
                Point{I};
              }
            };

    NEWI[i] = tmp[0]; i++;
  EndFor

  // Spline the extruded points
  IL = newl;

  If ( !(DIR1[0] == 0 && DIR1[1] == 0 && DIR1[2] == 0) &&
       !(DIR2[0] == 0 && DIR2[1] == 0 && DIR2[2] == 0)   )
    Spline(IL) = {IB1EX, NEWI[0] : NEWI[i-1], IB2EX};
  EndIf

  If ( !(DIR1[0] == 0 && DIR1[1] == 0 && DIR1[2] == 0) &&
        (DIR2[0] == 0 && DIR2[1] == 0 && DIR2[2] == 0)   )
    Spline(IL) = {IB1EX, NEWI[0] : NEWI[i-1]};
  EndIf

  If (  (DIR1[0] == 0 && DIR1[1] == 0 && DIR1[2] == 0) &&
       !(DIR2[0] == 0 && DIR2[1] == 0 && DIR2[2] == 0)   )
    Spline(IL) = {NEWI[0] : NEWI[i-1], IB2EX};
  EndIf

  If ( (DIR1[0] == 0 && DIR1[1] == 0 && DIR1[2] == 0) &&
       (DIR2[0] == 0 && DIR2[1] == 0 && DIR2[2] == 0)   )
    Spline(IL) = {NEWI[0] : NEWI[i-1]};
  EndIf

  Geometry.AutoCoherence = 1; Coherence;

Return
